// Tree.java
// 
// Copyright (C) 1999-2001 Washington University School of Medicine
// and Howard Hughes Medical Institute
// All rights reserved
//
// Created: 1999
// Author: Christian M. Zmasek
// zmasek@genetics.wustl.edu
// http://www.genetics.wustl.edu/eddy/people/zmasek/



// Internally, the trees are -- after they are completely constructed --
// always rooted binary trees.


package forester.tree;


import java.io.*;
import java.util.Stack;
import java.util.Vector;
import java.util.StringTokenizer;
import java.net.*;
import java.util.Hashtable;  // For JDK 1.1.



/**

@author Christian M. Zmasek

@version 1.710 -- last modified: 05/26/01

*/
public class Tree implements Serializable {

    static final int MAX_LENGTH = 10;
    // Max length for seq names, when printing out "clean" NH.
    
    static final long serialVersionUID = -6847390332113L;
    // See Horstmann and Cornell: Core Java 1.2 Vol I, p.686.

    private Node    ext_node0              = null,
                    root                   = null;
    private double  highest_lnL            = 0.0,
                    lowest_lnL             = 0.0,
                    real_height            = 0.0;
                    
    private int     external_nodes         = 0,
                    number_of_duplications = -1;
                    
    private boolean rooted                     = false,
                    more_than_bin_nodes_in_NH  = true;

    private String  name = "";

    private Hashtable idhash = null;

    private int j = 0; // Used in recursive method "preorderReID".

    
    /**

    Default Tree constructor.
    Constructs empty Tree.
    
    */
    public Tree() {
        setExtNode0( null );
        setRoot( null );
    }



    /**

    Tree constructor which constructs a Tree from the nh_string.
    The tree is considered unrooted if the deepest node has more than
    two children, otherwise it is considered rooted.

    @param nh_string String in New Hampshire (NH) or New Hampshire X (NHX) format

    */
    public Tree( String nh_string ) throws Exception {
        
        idhash = null;
        setRooted( false );
        
        boolean      first = true;
        int          i;
        String       internal_node_info, A, B, next;
        StringBuffer sb;
        Stack        st;

        // To simulate >binary trees.
        String null_distance = new String( ":" + Node.DISTANCE_NULL ); 

        nh_string = TreeHelper.removeWhiteSpace( nh_string );
       
        nh_string = TreeHelper.removeComments( nh_string );

        // Remove anything before first "(", unless tree is just one node.
        if ( !TreeHelper.isEmpty( nh_string ) && nh_string.charAt( 0 ) != '('
        && nh_string.indexOf( "(" ) != -1 ) {
            int x = nh_string.indexOf( "(" );
            nh_string = nh_string.substring( x );
        }

        // If ';' at end, remove it.
        if ( !TreeHelper.isEmpty( nh_string )
        && nh_string.endsWith( ";" ) ) {
            nh_string = nh_string.substring( 0, nh_string.length() - 1 );
        }

        if ( TreeHelper.countAndCheckParantheses( nh_string ) <= -1 ) {
            String message = "Tree: Tree( String ): Error in NHX format: ";
            message += "open parantheses != close parantheses.";
            throw new Exception( message );
        }
        if ( !TreeHelper.checkCommas( nh_string ) ) {
            String message = "Tree: Tree( String ): Error in NHX format: ";
            message += "Commas not properly set.";
            throw new Exception( message );
        }
  
        // Conversion from nh string to tree object.

        // Empty Tree.
        if ( nh_string.length() < 1 ) {
            setExtNode0( null );
            setRoot( null );
        }

        // Check whether nh string represents a tree with more than
        // node or just a single node.
        // More than one node.
        else if ( nh_string.indexOf( "(" ) != -1 ) {

            A = B = next = "";

            st = new Stack();

            i = 0;

            while ( i <= nh_string.length() - 1 ) {
                if ( nh_string.charAt( i ) == ',' ) {
                    st.push( "," );
                }
                if ( nh_string.charAt( i ) == '(' ) {
                    st.push( "(" );
                }
                if ( nh_string.charAt( i ) != ','
                && nh_string.charAt( i ) != '('
                && nh_string.charAt( i ) != ')' ) {
                    sb = new StringBuffer( "" );
                    while ( i <= nh_string.length() - 1
                    && nh_string.charAt( i ) != ')'
                    && nh_string.charAt( i ) != ','   ) {
                        sb.append( nh_string.charAt( i ) );
                        i++;
                    }
                    i--;
                    st.push( sb.toString() );
                }

                // A ")" calls for connection of one kind or another.
                if ( nh_string.charAt( i ) == ')' ) {
                    internal_node_info = "";
                    // If present, read information for internal node.
                    if ( i <= nh_string.length() - 2
                    && nh_string.charAt( i + 1 ) != ')'
                    && nh_string.charAt( i + 1 ) != ',' ) {
                        i++;
                        sb = new StringBuffer( "" );
                        while ( i <= nh_string.length() - 1
                        && nh_string.charAt( i ) != ')'
                        && nh_string.charAt( i ) != ',' ) {
                            sb.append( nh_string.charAt( i ) );
                            i++;
                        }
                        i--;
                        internal_node_info = sb.toString();
                    }

                    first = true;

                    // Parsing between two parantheses.
                    do {

                        A = st.pop().toString();

                        if ( st.empty() ) {
                            connectInternal( internal_node_info );
                            break;
                        }

                        B = st.pop().toString();

                        if ( st.empty() ) {
                            if ( A.equals( "(" ) ) {
                                connectInternal( internal_node_info );
                                st.push( "(" );
                            }
                            else if ( A.equals( "," ) ) {
                                connectInternal( internal_node_info );
                                break;
                            }
                            else {
                                addNodeAndConnect( A, internal_node_info );
                                break;
                            }
                        }
                        else {
                            next = st.peek().toString();

                            if ( !next.equals( "(" ) && B.equals( "," )
                            && !A.equals( "," ) ) {
                                if ( first && !next.equals( "," ) ) {
                                    addNode( A );
                                    st.push( "," );
                                }
                                else {
                                    addNodeAndConnect( A, null_distance );
                                }
                                first = false;
                            }

                            else {
                                first = false;

                                if ( next.equals( "," ) && !B.equals( "," )
                                && !B.equals( "(" )
                                && A.equals( "," ) ) {
                                    addNodeAndConnect( B, null_distance );
                                }

                                else if ( !next.equals( "(" )
                                && B.equals( "," ) && A.equals( "," ) ) {
                                    connectInternal( null_distance );
                                    st.push( "," );
                                }

                                else if ( next.equals( "(" ) && B.equals( "," )
                                && !A.equals( "(" ) && !A.equals( "," ) ) {
                                    addNodeAndConnect( A, internal_node_info );
                                    st.pop();
                                    break;
                                }

                                else if ( next.equals( "(" ) && !B.equals( "(" )
                                && !B.equals( "," ) && A.equals( "," ) ) {
                                    addNodeAndConnect( B, internal_node_info );
                                    st.pop();
                                    break;
                                }

                                else if ( next.equals( "(" )
                                && B.equals( "," ) && A.equals( "," ) ) {
                                    connectInternal( null_distance );
                                    connectInternal( internal_node_info );
                                    st.pop();
                                    break;
                                }

                                else if ( next.equals( "," ) && B.equals( "(" )
                                && !A.equals( "(" ) && !A.equals( "," ) ) {
                                    addNodeAndConnect( A, internal_node_info );
                                    break;
                                }

                                else if ( next.equals( "," )
                                && B.equals( "(" ) && A.equals( "," ) ) {
                                    connectInternal( internal_node_info );
                                    break;
                                }

                                else if ( next.equals( "(" )
                                && B.equals( "(" ) && !A.equals( "(" ) ) {
                                    if ( A.equals( "," ) ) {
                                        connectInternal( internal_node_info );
                                    }
                                    else {
                                        addNodeAndConnect( A,internal_node_info) ;
                                    }
                                    break;
                                }

                                else if ( A.equals("(")
                                && ( ( next.equals("(") && B.equals(",") )
                                || ( next.equals(",") && B.equals("(") )
                                || ( next.equals("(") && B.equals("(") ) ) ) {
                                    connectInternal( internal_node_info );
                                    st.push( "(" );
                                    break;
                                }

                            }

                        } // end of else (st is not empty).

                    } while ( true );


                }

                i++;

            } // End of while loop going through nh_string.

            if ( !getExtNode0().getRoot().getChild1().isPseudoNode() 
            && !getExtNode0().getRoot().getChild2().isPseudoNode() ) {
                setRooted( true );  
            }
        }

        // Just one node.
        // Conversion from nh string to tree object with one node.
        else {
            addNode( nh_string );
            setRooted( true );
        }


        if ( !isEmpty() ) {
            setRoot( getExtNode0().getRoot() );
            if ( !isRooted() ) {
                getRoot().deleteData();
            }
            findExtremeLnL();
            calculateRealHeight();

            // Checks whether branch lenghts are actually bootstraps.
            // (In case of trees coming form NH format.)
            if ( areBranchLenghtsBootstraps() ) {
                moveBranchLenghtsToBootstrap();
            }
        }
    } // End of Tree constructor.



    /**

    Checks whether a Tree object is deleted (or empty).

    @return true if the tree is deleted (or empty), false otherwise

    */
    public boolean isEmpty() {
        return ( getExtNode0() == null );
    }



    /**

    Deletes this Tree.

    */
    public void delete() {
        setExtNode0( null );
        setRoot( null );
        highest_lnL            = 0.0;
        lowest_lnL             = 0.0;
        real_height            = 0.0;
        external_nodes         = 0;
        number_of_duplications = -1;
        name                   = "";
        idhash                 = null;
    }
 


    /**

    Returns all orthologs of the external Node n of this Tree. Orthologs are
    returned as Vector of references to Nodes. This tree must be binary
    and rooted, and speciation vs. duplication needs to be assigned for each of
    its internal Nodes. Returns null if this Tree is empty or if n is internal.
    (Last modified: 11/22/00)

    @param n external Node whose orthologs are to be returned

    @return Vector of references to all orthologous Nodes of Node n of this
            Tree, null if this Tree is empty or if n is internal 

    */
    public Vector getOrthologs( Node n ) {

        Node   node  = n,
               prev  = null;
        Vector v     = new Vector();

        if ( !node.isExternal() || isEmpty() ) {
            return null;
        }

        while ( !node.isRoot() ) {
            prev = node;
            node = node.getParent();
            if ( !node.isDuplication() ) {
                if ( node.getChild1() == prev ) {
                    addIntoVector( node.getChild2().getAllExternalChildren(), v );
                }
                else {
                    addIntoVector( node.getChild1().getAllExternalChildren(), v );
                }            
            }
        }

        v.trimToSize();

        return v;
    }



    // Adds all element from Vector from into Vector addto.
    // (Last modified: 11/22/00)
    private void addIntoVector( Vector from, Vector addto ) {
        for ( int i = 0; i < from.size(); ++i ) {
            addto.addElement( from.elementAt( i ) );
        }
    }



    /**

    Returns all Nodes which are connected to external Node n of this Tree by
    a path containing only speciation events. Nodes are returned as Vector of
    references to Nodes. This tree must be binary and rooted, and speciation
    vs. duplication needs to be assigned for each of its internal Nodes.
    Returns null if this Tree is empty or if n is internal.
    (Last modified: 11/22/00)

    @param n external Node whose strictly speciation related Nodes are to be 
             returned 

    @return Vector of references to all strictly speciation related Nodes of
            Node n of this Tree, null if this Tree is empty or if n is internal

    */
    public Vector getStrictlySrelatedNodes( Node n ) {

        Node node    = n,
             deepest = null;
        Vector v     = new Vector();

        if ( !node.isExternal() || isEmpty() ) {
            return null;
        }

        while ( !node.isRoot() && !node.getParent().isDuplication() ) {
            node = node.getParent();
        }

        deepest = node;
        deepest.setIndicatorsToZero();

        do {
            if ( !node.isExternal() ) {
                if ( node.getIndicator() == 0 ) {
                    node.setIndicator( 1 );
                    if ( !node.isDuplication() ) {
                        node = node.getChild1();
                    }
                }
                if ( node.getIndicator() == 1 ) {
                    node.setIndicator( 2 );
                    if ( !node.isDuplication() ) {
                        node = node.getChild2();
                    }
                }
                if ( node != deepest && node.getIndicator() == 2 ) {
                    node = node.getParent();
                }
            }
            else {
                if ( node != n ) {
                    v.addElement( node );
                }
                if ( node != deepest ) {
                    node = node.getParent();
                }
                else {
                    node.setIndicator( 2 );
                }
            }
        } while ( node != deepest || deepest.getIndicator() != 2 );

        v.trimToSize();

        return v;
    }



    /**

    Returns the sequence names of all external Nodes of this Tree
    as array of Strings. 
    (Last modified: 11/13/00)

    @return all external sequence names as String[]

    */
    public String[] getAllExternalSeqNames() {
        if ( isEmpty() ) {
            return null;
        }
        int i = 0;
        String[] names = new String[ getNumberOfExtNodes() ];
        Node node = getExtNode0();
        while ( node != null ) {
                names[ i++ ] = new String( node.getSeqName() );
                node = node.getNextExtNode();
        } 
        return names;
    }



    /**

    Removes external Node from this Tree.
    If this tree has only one node, a empty tree will be returned.

    @param n Node to remove

    */
    public void removeExtNode( Node n ) {

        if ( isEmpty() ) {
            return;
        }
     
        // Returns an empty tree.   
        if ( getNumberOfExtNodes() == 1 && n == getExtNode0() ) {
            setExtNode0( null );
            setRoot( null );
            highest_lnL            = 0.0;
            lowest_lnL             = 0.0;
            real_height            = 0.0;
            external_nodes         = 0;
            number_of_duplications = -1;
            idhash                 = null;
            return;  
        }

        Node removed_node = null;

        removed_node = n;

        if ( removed_node == getExtNode0() ) {
            setExtNode0( removed_node.getNextExtNode() );
            removed_node.getNextExtNode().setPrevExtNode( null );
        }
        else if ( removed_node.getNextExtNode() == null ) {
            removed_node.getPrevExtNode().setNextExtNode( null );
        }
        else {
            removed_node.getNextExtNode().setPrevExtNode( removed_node.getPrevExtNode() );
            removed_node.getPrevExtNode().setNextExtNode( removed_node.getNextExtNode() );
        }


        Node p = removed_node.getParent();

        if ( p.isRoot() ) {
            if ( removed_node.isChild1() ) {
                setRoot( getRoot().getChild2() );
                getRoot().setParent( null );
            }
            else {
                setRoot( getRoot().getChild1() );
                getRoot().setParent( null );
            }
        }
        else {
            Node pp = removed_node.getParent().getParent();
            if ( p.isChild1() ) {
                if ( removed_node.isChild1() ) {
                    p.getChild2().setDistanceToParent( addDist( p.getDistanceToParent(), p.getChild2().getDistanceToParent() ) );
                    pp.setChild1( p.getChild2() );
                    p.getChild2().setParent( pp );
                }
                else {
                    p.getChild1().setDistanceToParent( addDist( p.getDistanceToParent(), p.getChild1().getDistanceToParent() ) );
                    pp.setChild1( p.getChild1() );
                    p.getChild1().setParent( pp );
                }

            }
            else {
                if ( removed_node.isChild1() ) {
                    p.getChild2().setDistanceToParent( addDist( p.getDistanceToParent(), p.getChild2().getDistanceToParent() ) );
                    pp.setChild2( p.getChild2() );
                    p.getChild2().setParent( pp );
                }
                else {
                    p.getChild1().setDistanceToParent( addDist( p.getDistanceToParent(), p.getChild1().getDistanceToParent() ) );
                    pp.setChild2( p.getChild1() );
                    p.getChild1().setParent( pp );
                }
            }

            while ( pp != getRoot() ) {
                pp.setSumExtNodes( pp.getSumExtNodes() - 1 );
                pp = pp.getParent();
            }

            pp.setSumExtNodes( pp.getSumExtNodes() - 1 );




        }
        setExtNodes( getNumberOfExtNodes() - 1 );
        idhash = null;
        return;
    }


    /**

    Private helper method for removeExtNode( Node ).

    */
    private double addDist( double a, double b ) {
        if ( a >= 0.0 && b >= 0.0 ) {
            return a + b;
        }
        else if ( a >= 0.0 ) {
            return a;
        }
        else if ( b >= 0.0 ) {
            return b;
        }
        else if ( a == Node.DISTANCE_NULL && b == Node.DISTANCE_NULL ) {
            return Node.DISTANCE_NULL;
        }
        return Node.DISTANCE_DEFAULT;
    }



    /**

    Returns a Vector with references to all Nodes of this Tree which
    have a matching sequence name.
    
    @param seqname Sequence name (String) of Nodes to find

    @return Vector of references to Nodes of this Tree with matching
    sequence names

    @see #getNodesWithMatchingSpecies(String)

    */
    public Vector getNodes( String seqname ) throws Exception {
        if ( isEmpty() ) {
            return null;
        }
        seqname = seqname.trim();
        Vector nodes = new Vector();

        PreorderTreeIterator i = new PreorderTreeIterator( this );
        while ( !i.isDone() ) {
            if ( i.currentNode().getSeqName().trim().equals( seqname ) ) {
                nodes.addElement( i.currentNode() );
            }
            i.next();
        }

        nodes.trimToSize();
        return nodes;
    }



    /**

    Returns a Node of this Tree which has a matching sequence name seqname.
    Throws an Exception if seqname is not present in this or not unique.    
    (Last modifed: 12/03/00)

    @param seqname Sequence name (String) of Node to find

    @return Node with matchin sequence name 

    */
    public Node getNode( String seqname ) throws Exception {
        if ( isEmpty() ) {
            return null;
        }
        seqname = seqname.trim();
        Vector nodes = getNodes( seqname );
        
        if ( nodes == null || nodes.size() < 1 ) {
            throw new Exception( seqname + " not found" );
        }
        if ( nodes.size() > 1 ) {
            throw new Exception( seqname + " not unique" );
        }
        return ( Node ) nodes.elementAt( 0 );
    }

    /**

    Returns a Vector with references to all Nodes of this Tree which
    have a matching species name.

    @param specname species name (String) of Nodes to find

    @return Vector of references to Nodes of this Tree with matching
    species names.

    @see #getNodes(String)

    */
    public Vector getNodesWithMatchingSpecies( String specname ) throws Exception {
        if ( isEmpty() ) {
            return null;
        }
        specname = specname.trim();
        Vector nodes = new Vector();

        PreorderTreeIterator i = new PreorderTreeIterator( this );
        while ( !i.isDone() ) {
            if ( i.currentNode().getSpecies().trim().equals( specname ) ) {
                nodes.addElement( i.currentNode() );
            }
            i.next();
        }

        nodes.trimToSize();
        return nodes;
    }



    /**

    Finds the Node of this Tree which has a matching ID number.
    Takes O(n) time. After method hashIDs() has been called it runs in
    constant time.
    (Last modified: 11/20/00)

    @param id ID number (int) of the Node to find

    @return Node with matching ID, null if Node not found

    */
    public Node getNode( int id ) {

        if ( isEmpty() ) {
            return null;
        }

        if ( idhash != null ) {
            return ( Node ) idhash.get( new Integer( id ) );
        }
        else {
            return getNodeHelper( getRoot(), id );
        }
    }  



    // Helper method for "getNode(int)".
    // Searches for id by traversing tree recusively.
    // (Last modified: 11/18/00)
    private Node getNodeHelper( Node n, int id ) {

       if ( n == null || n.getID() == id ) {
           return n;
       }
       else {
           Node n1 = getNodeHelper( n.getChild1(), id ),
                n2 = getNodeHelper( n.getChild2(), id );
           if ( n1 != null ) {
              return n1;
           }
           return n2;
       }
    }



    /**

    Swaps the the two childern of a Node with ID id of this Tree.
    
    @param id ID (int) of Node
    
    */
    public void swapChildren( int id ) {
        swapChildren( getNode( id ) );
    }



    /**

    Swaps the the two childern of a Node node of this Tree.
    
    @param node a Node of this Tree
    
    */
    public void swapChildren( Node node ) {
        if ( isEmpty() ) {
            return;
        }
        if ( node.isExternal() ) {
            return;
        }
        Node temp = node.getChild1();
        node.setChild1( node.getChild2() );
        node.setChild2( temp );
        adjustExtNode0();
        adjustReferencesInExtNodes();
        adjustNodeCount();
        recalculateAndReset();
    }



    /**

    Returns the subtree of this Tree which has the Node with ID id
    as its root Node.
    
    @param id ID (int) of Node

    */
    public Tree subTree( int id ) throws Exception {
        if ( isEmpty() ) {
            return null;
        }
        Tree tree = copyTree();
        Node node = tree.getNode( id );
        if ( node.isExternal() ) {
            String message = "Tree: subTree( int ): ";
            message += "Can not get a sub tree from a external node.";
            throw new Exception( message );
        }
        node.setParent( null );
        node.setDistanceToParent( Node.DISTANCE_DEFAULT );
        tree.setRooted( true );
        tree.setRoot( node );
        tree.adjustExtNode0();
        tree.adjustReferencesInExtNodes();
        tree.adjustNodeCount();
        tree.recalculateAndReset();
        tree.setExtNodes( tree.getRoot().getSumExtNodes() );
        return tree;
    }



    /**

    Places the root of this Tree on the parent branch of the
    Node with a corresponding ID. The new root is always placed
    on the middle of the branch.
    If the resulting reRooted Tree is to be used any further,
    in most cases the following three methods have to be
    called on the resulting Tree:
    adjustNodeCount()
    recalculateAndReset()
    setExtNodes( getRoot().getSumExtNodes() ) 

    @param id ID (int) of Node of this Tree
    
    */
    public void reRoot( int id ) throws Exception {
        reRoot( getNode( id ), "" );
    }

    
    
    /**

    Places the root of this Tree on the parent branch Node n.
    The new root is always placed on the middle of the branch.
    If the resulting reRooted Tree is to be used any further,
    in most cases the following three methods have to be
    called on the resulting Tree:
    adjustNodeCount()
    recalculateAndReset()
    setExtNodes( getRoot().getSumExtNodes() ) 

    @param n Node of this Tree 
    
    */ 
    public void reRoot( Node n ) throws Exception {
        reRoot( n, "" );
    }



    /**

    Places the root of this Tree on the parent branch of the
    Node with a corresponding ID. Information for the new root Node
    is provided in a String using NHX tags. The new root is always placed
    on the middle of the branch.
    If the resulting reRooted Tree is to be used any further,
    in most cases the following three methods have to be
    called on the resulting Tree:
    adjustNodeCount()
    recalculateAndReset()
    setExtNodes( getRoot().getSumExtNodes() ) 

    @param id ID (int) of Node of this Tree
    @param s  String describing new Node
    
    */ 
    public void reRoot( int id, String s ) throws Exception {
        reRoot( getNode( id ), s );
        
    }
    


    /**

    Places the root of this Tree on the parent branch Node n.
    Information for the new root Node is provided in a String
    using NHX tags. The new root is always placed on the middle
    of the branch.
    If the resulting reRooted Tree is to be used any further,
    in most cases the following three methods have to be
    called on the resulting Tree:
    adjustNodeCount()
    recalculateAndReset()
    setExtNodes( getRoot().getSumExtNodes() ) 
    <br>
    last modified: 05/22/01

    @param n Node of this Tree 
    @param s String describing new Node
    
    */ 
    public void reRoot( Node n, String s ) throws Exception {

        if ( isEmpty() ) {
            return;
        }

        Node nodeA = n,
             node  = null;

        double  distance1,
                distance2;
        float   lnL1,
                lnL2;

        boolean sign_worse1,
                sign_worse2,
                lnL_assign1,
                lnL_assign2;

        int     bootstrap1,
                bootstrap2;

        if ( nodeA.isPseudoNode() ) {
            String message = "Tree: reRoot: ";
            message += "Can not place root in >bin node.";
            throw new Exception( message );
        }

        setRooted( true );

        if ( nodeA.isRoot() ) {
            return;
        }


        if ( nodeA.getParent().isRoot() ) {
            if ( getRoot().getChild1().isPseudoNode()
            || getRoot().getChild2().isPseudoNode() ) {
                if ( getRoot().getChild1() == nodeA
                &&   getRoot().getChild2().isPseudoNode() ) {
                    node = getRoot().getChild2();
                }
                if ( getRoot().getChild2() == nodeA
                &&   getRoot().getChild1().isPseudoNode() ) {
                    node = getRoot().getChild1();
                }
                if ( nodeA.getDistanceToParent() == Node.DISTANCE_DEFAULT ) {
                    node.setDistanceToParent( Node.DISTANCE_DEFAULT );
                }
                else {
                    node.setDistanceToParent( 
                     nodeA.getDistanceToParent() / 2 );
                    nodeA.setDistanceToParent(
                     nodeA.getDistanceToParent() / 2 );
                }

                node.setLnLonParentBranch( nodeA.getLnLonParentBranch() );
                node.setSignificantlyWorse( nodeA.significantlyWorse() );
                node.setLnLonParentBranchAssigned(
                 nodeA.isLnLonParentBranchAssigned() );
                node.setBootstrap( nodeA.getBootstrap() ); 
            }
            return;
        }

        Node nodeB = nodeA.getParent();
        Node nodeC = nodeB.getParent();

        // New Root.
        Node new_root = new Node( s );
        new_root.setParent( null );
        if ( nodeB.getChild2() == nodeA ) {
            new_root.setChild1( nodeA );
            new_root.setChild2( nodeB );
        }
        else {
            new_root.setChild1( nodeB );
            new_root.setChild2( nodeA );
        }

        distance1   = nodeC.getDistanceToParent();
        lnL1        = nodeC.getLnLonParentBranch();
        sign_worse1 = nodeC.significantlyWorse();
        lnL_assign1 = nodeC.isLnLonParentBranchAssigned();
        bootstrap1  = nodeC.getBootstrap();

        nodeC.setDistanceToParent( nodeB.getDistanceToParent() );

        nodeC.setLnLonParentBranch( nodeB.getLnLonParentBranch() );
        nodeC.setLnLonParentBranchAssigned( nodeB.isLnLonParentBranchAssigned() );
        nodeC.setSignificantlyWorse( nodeB.significantlyWorse() );
        nodeC.setBootstrap( nodeB.getBootstrap() );      


        nodeB.setLnLonParentBranch( nodeA.getLnLonParentBranch() );
        nodeB.setLnLonParentBranchAssigned( nodeA.isLnLonParentBranchAssigned() );
        nodeB.setSignificantlyWorse( nodeA.significantlyWorse() );
        nodeB.setBootstrap( nodeA.getBootstrap() ); 


        // New root is always placed in the middle of the branch:
        if ( nodeA.getDistanceToParent() == Node.DISTANCE_DEFAULT ) {
            nodeB.setDistanceToParent( Node.DISTANCE_DEFAULT );
        }
        else {
            nodeB.setDistanceToParent( nodeA.getDistanceToParent() / 2 );
            nodeA.setDistanceToParent( nodeA.getDistanceToParent() / 2 );
        }

        // nodeA:
        nodeA.setParent( new_root );


        // nodeB:
        if ( nodeB.getChild1() == nodeA ) {
            nodeB.setChild1( nodeC );
        }
        else {
            nodeB.setChild2( nodeC );
        }
        nodeB.setParent( new_root );

        // moving to the old root, swapping references:
        while ( !nodeC.isRoot() ) {

            nodeA = nodeB;
            nodeB = nodeC;
            nodeC = nodeC.getParent();

            if ( nodeB.getChild1() == nodeA ) {
                nodeB.setChild1( nodeC );
            }
            else {
                nodeB.setChild2( nodeC );
            }
            nodeB.setParent( nodeA );

            distance2   = nodeC.getDistanceToParent();
            lnL2        = nodeC.getLnLonParentBranch();
            sign_worse2 = nodeC.significantlyWorse();
            lnL_assign2 = nodeC.isLnLonParentBranchAssigned();
            bootstrap2  = nodeC.getBootstrap();

            nodeC.setDistanceToParent( distance1 );
            nodeC.setLnLonParentBranch( lnL1 );
            nodeC.setSignificantlyWorse( sign_worse1 );
            nodeC.setLnLonParentBranchAssigned( lnL_assign1 );
            nodeC.setBootstrap( bootstrap1 );

            distance1   = distance2;
            lnL1        = lnL2;
            sign_worse1 = sign_worse2;
            lnL_assign1 = lnL_assign2;
            bootstrap1  = bootstrap2;
        }


        // removing the old root:
        if ( nodeC.getChild1() == nodeB ) {
            node = nodeC.getChild2();
        }
        else {
            node = nodeC.getChild1();
        }
        node.setParent( nodeB );

        if ( nodeC.getDistanceToParent() == Node.DISTANCE_NULL
        && node.getDistanceToParent() == Node.DISTANCE_NULL ) {
            node.setDistanceToParent( Node.DISTANCE_NULL );
        }
        else if ( ( nodeC.getDistanceToParent() == Node.DISTANCE_DEFAULT
        ||  nodeC.getDistanceToParent() == Node.DISTANCE_NULL )
        && ( node.getDistanceToParent() == Node.DISTANCE_DEFAULT
        ||  node.getDistanceToParent() == Node.DISTANCE_NULL ) ) {
            node.setDistanceToParent( Node.DISTANCE_DEFAULT );
        }
        else {
            node.setDistanceToParent( ( nodeC.getDistanceToParent() >= 0.0 ?
            nodeC.getDistanceToParent() : 0.0 )
            + ( node.getDistanceToParent() >= 0.0 ?
            node.getDistanceToParent() : 0.0 ) );
        }


        if ( nodeC.getDistanceToParent() != Node.DISTANCE_NULL ) {
            node.setLnLonParentBranch( nodeC.getLnLonParentBranch() );
            node.setSignificantlyWorse( nodeC.significantlyWorse() );
            node.setLnLonParentBranchAssigned( nodeC.isLnLonParentBranchAssigned() );
        }

        node.setBootstrap( nodeC.getBootstrap() );

        if ( nodeB.getChild1() != nodeC ) {
            nodeB.setChild2( node );
        }
        else {
            nodeB.setChild1( node );
        }


        setRoot( new_root );

        adjustExtNode0();
        adjustReferencesInExtNodes();
        return;

    }



    /**

    Modifies this Tree with the branch lenghts of Tree t.
    Important (but obvious): The topology of both trees
    needs to be the same. (The method is not robust in this point, and will
    produce wrong results if the internal topology differs.)
    Furthermore, the combination of sequence name, species, and EC number
    needs to be unique for each external node.
    <br>
    Last modified: 05/22/01 

    @param t the Tree to copy the branch lenghts from

    */
    public void copyBranchLengthValuesFrom( Tree t ) throws Exception {
        if ( isEmpty() || t.isEmpty() ) {
            return;
        }
        Node node0, node2_, node1_;
        Node[] nodes1, nodes2;
        int i = 0;
        setIndicatorsToZero();
        t.setIndicatorsToZero();

        // Visit each node of this and set distance to parent to 0.0:
        Node node2, node1 = getExtNode0();
        while ( node1 != null ) {
            node2 = node1;
            while ( node2 != null ) {
                if ( !node2.isExternal() ) {
                    node2.setDistanceToParent( 0.0 );
                }
                node2 = node2.getParent();
            }
            node1 = node1.getNextExtNode();
        }

        // Visit each internal node of t:
        node1 = t.getExtNode0();
        while ( node1 != null ) {
            node2 = node1;
            while ( node2 != null ) {
                if ( node2.getIndicator() == 0
                && !node2.isPseudoNode() ) {
                    node2.setIndicator( 1 );
                    nodes2 = node2.copyAllExtChildren();
                    node0 = nodes2[ 0 ];
                    i++;
                    node1_ = getExtNode0();
                    // Visit each internal node of this:
                    while ( node1_ != null ) {
                        node2_ = node1_;
                        if ( node0.equals( node1_ ) ) {
                            while ( node2_ != null ) {
                                if ( node2_.getIndicator()
                                != i && node2_.getIndicator() != -1 ) {
                                    node2_.setIndicator( i );
                                    if ( nodes2.length
                                    == node2_.getSumExtNodes() ) {
                                        nodes1 = node2_.copyAllExtChildren();
                                        if ( Node.compareArraysOfNodes(
                                         nodes2, nodes1 ) ) {
                                            node2_.setIndicator( -1 );
                                            node2_.setDistanceToParent(
                                             node2.getDistanceToParent() );
                                        }
                                        else {
                                            String message = "Trees were ";
                                            message += "not identical.";
                                            throw new Exception( message );
                                        }
                                    }
                                }
                                node2_ = node2_.getParent();
                            }
                        }
                        node1_ = node1_.getNextExtNode();
                    } //end of visiting each internal node of this.
                }
                node2 = node2.getParent();
            }
            node1 = node1.getNextExtNode();
        }
        if ( !t.isRooted() ) {
            setRooted( false );
        }
        else {
            setRooted( true );
        }
        recalculateAndReset();
        reassignIDs();

        return;
    }



    /**

    Recalculates and resets parameters of this Tree:
    highest and lowest lnL, real height, sum of ext Nodes. 
    To be used after Tree has been modified.
    (Last modified: 11/18/00)

    */
    public void recalculateAndReset() {

        if ( isEmpty() ) {
            return;
        }

        findExtremeLnL();
        calculateRealHeight();
        setExtNodes( getRoot().getSumExtNodes() );
    }
   
   
    
    /**

    Returns a deep copy of this Tree.
    (Last modified: 11/18/00)

    */
    public Tree copyTree() {

        Tree tree                       = new Tree();

        if ( isEmpty() ) {
            tree.delete();
            return tree;
        }

        tree.rooted                     = rooted;
        tree.highest_lnL                = highest_lnL;
        tree.lowest_lnL                 = lowest_lnL;
        tree.real_height                = real_height;
        tree.external_nodes             = external_nodes;
        tree.number_of_duplications     = number_of_duplications;
        tree.more_than_bin_nodes_in_NH  = more_than_bin_nodes_in_NH;
        tree.name                       = new String( name );
        tree.idhash                     = idhash;

        tree.root = Node.copyTree( root );

        tree.getRoot().setParents();

        tree.adjustExtNode0();
        tree.adjustReferencesInExtNodes();
        tree.setIndicatorsToZero();
        return tree;
    }


   
    /**
    
    Removes the root Node of this Tree 
    and makes at least a trifurcation at its basal node.

    */
    public void unRootAndTrifurcate() {
        if ( isEmpty() ) {
            return;
        }
        unRoot();
        double sum;
        if ( getNumberOfExtNodes() <= 2 ) {
            return;
        }

        sum = getRoot().getChild2().getDistanceToParent()
        + getRoot().getChild1().getDistanceToParent();
        if ( getRoot().getChild2().isExternal() ) {
            if ( sum >= 0.0 ) {
                getRoot().getChild2().setDistanceToParent( sum );
            }
            else {
                getRoot().getChild2().setDistanceToParent( Node.DISTANCE_DEFAULT );
            }
            getRoot().getChild1().setDistanceToParent( Node.DISTANCE_NULL );
        }
        else {
            if ( sum >= 0.0 ) {
                getRoot().getChild1().setDistanceToParent( sum );
            }
            else {
                getRoot().getChild1().setDistanceToParent( Node.DISTANCE_DEFAULT );
            }
            getRoot().getChild2().setDistanceToParent( Node.DISTANCE_NULL );
        }
        recalculateAndReset();
        return;
    }

    
    /**
    
    Removes the root Node this Tree.

    */
    public void unRoot() {
        if ( isEmpty() ) {
            return;
        }

        setIndicatorsToZero();
        if ( !isRooted() || getNumberOfExtNodes() <= 1 ) {
            return;
        }
        setRooted( false );
        recalculateAndReset();
        return;
    }



    /**

    Returns the root Node of this Tree.

    */
    public Node getRoot() {
        return root;
    }
    
    
    
    void setRoot( Node n ) {
        root = n;
    }

    
    
    /**

    Returns true is this Tree is rooted.

    */
    public boolean isRooted() {
        return rooted;
    }
    
    
    
    /**

    Sets whether this Tree is rooted or not.

    */
    public void setRooted( boolean b ) {
        rooted = b;
    }
    
    
    
    private void setHighestLnL( double d ) {
        highest_lnL = d;
    }
    
    
    
    /**

    Returns the highest log likelihood value associated with branches
    of this Tree (double).
    
    @see #findExtremeLnL()

    */
    public double getHighestLnL() {
        return highest_lnL;
    }
    
    
    
    private void setLowestLnL( double d ) {
        lowest_lnL = d;
    }
    
    
    
    /**
    
    Returns the lowest log likelihood value associated with branches
    of this Tree (double).

    @see #findExtremeLnL()

    */
    public double getLowestLnL() {
        return lowest_lnL;
    }
    
    
    
    /**

    Returns the real height of this Tree - the longest distance
    from root to external node. Is either calculated with
    "calculateRealHeight()" or manualy set with
    "setRealHeight(double)".

    @return the real height of this Tree - as calculated
            with "calculateRealHeight()" or set with
            "setRealHeight(double)"

    @see #calculateRealHeight()
    @see #setRealHeight(double)

    */
    public double getRealHeight () {
        return real_height ;
    }
    
    

    /**

    Sets the real height of this Tree - the longest distance
    from root to external node. Could either be calculated with
    "calculateRealHeight()" or manualy set with this method.
   
    @param the real height of this Tree

    @see #calculateRealHeight()

    */
    public void setRealHeight ( double d ) {
        real_height  = d;
    }



    /**

    Returns the number of duplications of this Tree (int).
    A return value of -1 indicates that the number of duplications
    is unknown.

    */
    public int getNumberOfDuplications() {
        return number_of_duplications;
    }
    
    
    
    /**
    
    Sets the number of duplications of this Tree (int).
    A value of -1 indicates that the number of duplications
    is unknown.
    
    @param clean_nh set to true for clean NH format
    
    */
    public void setNumberOfDuplications( int i ) {
        if ( i < 0 ) {
            number_of_duplications = -1;    
        }
        else {
            number_of_duplications = i;
        }
    }
    
    
    
    /**

    Returns the first external Node.

    */
    public Node getExtNode0() {
        return ext_node0;
    }
    
    
    
    private void setExtNode0( Node n ) {
        ext_node0 = n;
    }
    
    
    
    /**

    Returns the sum of external Nodes of this Tree (int).
  
    */
    public int getNumberOfExtNodes() {
        return external_nodes;
    }
    

    
    /**

    Sets the sum of external Nodes of this Tree (int).
  
    */
    public void setExtNodes( int i ) {
        external_nodes = i;
    }
    
    
    
    /**
    
    Returns whether to allow more than binary Nodes in New Hampshire (NH)
    output. The basal node can always appear trifurcated in the NH output.

    */
    private boolean allowMoreThanBinaryNodesInNHoutput() {
        return more_than_bin_nodes_in_NH;
    }


    
    
    /**

    Sets whether to allow more than binary Nodes in New Hampshire (NH)
    output. The basal node can always appear trifurcated in the NH output.

    */
    public void allowMoreThanBinaryNodesInNHoutput( boolean b ) {
        more_than_bin_nodes_in_NH = b;
    }



    /**
    
    Returns the name of this Tree.

    */
    public String getName() {
        return name;
    }


    
    
    /**

    Sets the name of this Tree to s.

    */
    public void setName( String s ) {
        name = s;
    }



    /**

    Prints descriptions of all external Nodes of this Tree to the console.

    */
    public void printExtNodes() {
        if ( isEmpty() ) { return; }
        Node node = getExtNode0();
        while ( node != null ) {
            System.out.println( node + "\n" );
            node = node.getNextExtNode();
        }
    }



    /**

    Prints descriptions of all Nodes of this Tree to the console.

    */
    public void printAllNodes() {
        if ( isEmpty() ) { return; }
        getRoot().preorderPrint();
    }



    /**

    Sets the indicators of all Nodes of this Tree to 0.

    */
    public void setIndicatorsToZero() {
        if ( isEmpty() ) {
            return;
        }
        getRoot().setIndicatorsToZero();
        
    }



    /**

    Calculates the real height of this (rooted) Tree -- the longest distance
    from root to external node.
    The height of this is stored and can be obtained with "getRealHeight()".
    Returns the difference in heights of the two subtrees originating at the
    root of this (which obviously only makes sense if this is rooted and binary
    at the root).
    Difference = height of subtree at child1 - height of subtree at child2
    (i.e. is positive if subtree 1 is higher, negative otherwise).
    <p>
    Takes into account collapsed Nodes.
    <p>
    (Last modified: 01/18/01)

    @return the difference in heigh of the two subtrees originating at the
            root of this Tree

    */
    public double calculateRealHeight() {

        if ( isEmpty() ) {
            setRealHeight( 0.0 );
            return 0.0;
        }

        Node   root   = getRoot();
        Node   child1 = root.getChild1(),
               child2 = root.getChild2();
        double r      = 0.0, 
               l      = 0.0,
               l1     = 0.0,          
               l2     = 0.0,
               diff   = 0.0;

        
        if ( isRooted() ) {
            r = root.getDistanceToParent();
            if ( r < 0.0 ) {
               r = 0.0;
            }
        }

        if ( child1 != null && child2 != null ) {
            l1 = calculateRealHeightHelper( child1 );
            l2 = calculateRealHeightHelper( child2 );

            diff = l1 - l2;

            l = ( l1 > l2 ? l1 : l2 );
        }
        
        setRealHeight( l + r );
        return diff;
    }



    private double calculateRealHeightHelper( Node n ) {
        if ( n == null || n.collapse() ) {
            return 0.0;
        }
        else {
            double l1 = calculateRealHeightHelper( n.getChild1() ), 
                   l2 = calculateRealHeightHelper( n.getChild2() ),
                   l  = n.getDistanceToParent();

            if ( l < 0.0 ) {
                l = 0.0;
            }
            return l + ( l1 > l2 ? l1 : l2 );
        }
    }



    /**

    Finds the highest of all log likelihood value associated with
    branches of this Tree.
    The result can be obtained with "getHighestLnL()". 
    (Last modifed: 11/18/00)    

    @see #getHighestLnL()

    */
    public void findExtremeLnL() {
        if ( isEmpty() ) {
            return;
        }
        lowest_lnL  = +Double.MAX_VALUE;
        highest_lnL = -Double.MAX_VALUE;
        findExtremeLnLHelper( getRoot() );
    }



    private void findExtremeLnLHelper( Node n ) {

        if ( n == null ) {
            return;
        }
        else {
            if ( n.isLnLonParentBranchAssigned()
            && !n.isPseudoNode() ) {
                if ( n.getLnLonParentBranch() > highest_lnL ) {
                    highest_lnL = n.getLnLonParentBranch();
                }
                if ( n.getLnLonParentBranch() < lowest_lnL ) {
                    lowest_lnL = n.getLnLonParentBranch();
                }
            }
            findExtremeLnLHelper( n.getChild1() );
            findExtremeLnLHelper( n.getChild2() );
        }
    }

  
  
    /**
  
    Returns whether this is a completely binary tree
    (i.e. all internal nodes are bifurcations).
    <br>
    (Last modifed: 05/26/01)
  
    */
    public boolean isCompletelyBinary() {
        if ( isEmpty() ) {
            return false;
        }
        return isCompletelyBinaryHelper( getRoot() );
    }
    
    
    // Helper for isCompletelyBinary()
    private boolean isCompletelyBinaryHelper( Node n ) {
        if ( n == null ) {
            return true;
        }
        else if ( n.isPseudoNode() ) {
            return false;
        }
        boolean b1 = isCompletelyBinaryHelper( n.getChild1() );
        boolean b2 = isCompletelyBinaryHelper( n.getChild2() );
        if ( !b1 || !b2 ) {
            return false;
        }
        else {
            return true;
        }  
    }
  
  
  
    /**
    
    Moves the values in the branch length field to the bootstrap field, for
    each Node of this Tree.
    Converts a Tree originating from a phylip treefile after bootstrapping
    and which therefore has its bootstrap values where the branch lenghts
    would be.
    
    */
    public void moveBranchLenghtsToBootstrap() {
        if ( isEmpty() ) {
            return;
        }
        setIndicatorsToZero();
        Node node2, node1 = getExtNode0();
        while ( node1 != null ) {
            node2 = node1;
            while ( node2 != null ) {
                if ( node2.getIndicator() == 0 ) {
                    if ( node2.getDistanceToParent() > 0.0 ) {
                        if ( !node2.isExternal() ) {
                            node2.setBootstrap( ( int ) node2.getDistanceToParent() );
                        }
                        node2.setDistanceToParent( Node.DISTANCE_DEFAULT );
                    }
                    else {
                        node2.setBootstrap( Node.BOOTSTRAP_DEFAULT );
                    }
                    node2.setIndicator( 1 );
                }
                node2 = node2.getParent();
            }
            node1 = node1.getNextExtNode();
        }
        recalculateAndReset();
    }



    /**

    Checks whether the branch length values actually are bootstrap values
    All external Nodes must have the same, >0, divisible by 10 branch length.
    Boostrap of root must be assigned.
    
    */
    public boolean areBranchLenghtsBootstraps() {
        if ( isEmpty() ) {
            return false;
        }
        if ( getNumberOfExtNodes() <= 1 ) {
            return false;
        }
        // Has bootstrap already been assigned?
        if ( getRoot().getBootstrap() != Node.BOOTSTRAP_DEFAULT ) {
            return false;
        }
        Node node = getExtNode0();
        double d1 = 0.0, d2 = 0.0;
        while ( node != null ) {
            d1 = node.getDistanceToParent();
            if ( d1 <= 0.0 || ( d1 %  10 ) != 0 ) {
                return false;
            }
            node = node.getNextExtNode();
            if ( node != null ) {
                d2 = node.getDistanceToParent();
                if ( d2 <= 0.0 || ( d2 %  10 ) != 0 ) {
                    return false;
                }
                if ( d1 != d2 ) {
                    return false;
                }
            }
        }
        return true;
    }



    /**
  
    Converts this Tree to a New Hampshire X (String) representation.
    
    @return New Hampshire X (String) representation of this
    
    @see #toNewHampshireX()
    
    */
    public String toString() {
        if ( isEmpty() ) {
            return "";
        }
        return toNewHampshireX();
    }



    /**
  
    Converts this Tree to a New Hampshire (String) representation.
    If the boolean clean_nh is true, the length of sequence names
    will be truncated to MAX_LENGTH (usually 10) characters and 
    everything after a "/" will be removed (including the "/").

    @param clean_nh set to true for clean NH format

    @return New Hampshire (String) representation of this
    
    */
    public String toNewHampshire( boolean clean_nh ) {

        if ( isEmpty() ) {
            return "";
        }
        int i = 0;
        String s = "", nh_string = "";
        Stack A_stack = new Stack();
        Stack B_stack = new Stack();
        Node node = root;
        
        setIndicatorsToZero();

        do {
            if ( !node.isExternal() ) {
                if ( node.getIndicator() == 0 ) {
                    node.setIndicator( 1 );
                    node = node.getChild1();
                }
                if ( node.getIndicator() == 1  ) {
                    node.setIndicator( 2 );
                    node = node.getChild2();
                }
                // Add internal node:
                if ( node.getIndicator() == 2 ) {

                    if ( !node.isPseudoNode()
                    || ( !allowMoreThanBinaryNodesInNHoutput()
                     && !node.getParent().isRoot()
                     && ( !(node.getParent().getChild1().isPseudoNode()  
                      && node.getParent().getChild2().isPseudoNode() )
                     || node.getParent().getChild1() == node ) ) ) {

                        i = 0;
                        while ( i < 2 * node.getSumExtNodes() - 2 ) {
                            B_stack.push( A_stack.pop() );
                            if ( !B_stack.peek().toString().equals( "(" )
                            && !B_stack.peek().toString().equals( ")" ) ) {
                                i++;
                            }
                        }
                        A_stack.push( "(" );
                        while ( !B_stack.empty() ) {
                            A_stack.push( B_stack.pop() );
                        }
                        A_stack.push( ")" );
                    }

                    if ( node.getDistanceToParent() >= 0.0
                    && !node.isPseudoNode() ) {
                        s = ":" + node.getDistanceToParent() + ",";
                    }
                    else if ( node.isPseudoNode()
                    && !allowMoreThanBinaryNodesInNHoutput()
                    && !node.getParent().isRoot()
                    && ( !(node.getParent().getChild1().isPseudoNode()
                     && node.getParent().getChild2().isPseudoNode() )
                     || node.getParent().getChild1() == node ) ) {
                        s = ":0.0,";
                    }
                    else {
                        s = ",";
                    }

                    A_stack.push( s );

                    node = node.getParent();
                }
            }
            // Add external node. Always preceeded by a "," which will need
            // to be removed in cases of "(,X"
            else {
                // If no seq_name is given, use species as name;
                // If no species is given, use EC as name
                // Otherwise, no information will be given.
                
                if ( clean_nh && !node.getSeqName().equals( "" ) ) {

                    String tmp = node.getSeqName();

                    try {
                        if ( tmp.length() > MAX_LENGTH ) {
                            tmp = tmp.substring( 0, MAX_LENGTH + 1 );
                        }
                        if ( tmp.indexOf( '/' ) > 0 ) {
                            tmp = tmp.substring( 0, tmp.indexOf( '/' ) );
                        }
                    }
                    catch ( Exception e ) {
                        System.err.println( "\ntoNewHampshire()" +
                        "Unexpected Exception.\n" );
                    }

                    s = "," + tmp;
                }
                else if ( !clean_nh && !node.getSeqName().equals( "" ) ) {
                    s = "," + node.getSeqName();
                }
                else if ( !clean_nh && !node.getSpecies().equals( "" ) ) {
                    s = "," + node.getSpecies();
                }
                else if ( !clean_nh && !node.getECnumber().equals( "" ) ) {
                    s = "," + node.getECnumber();
                }
                else {
                    s = ",\t";
                    // Introduces a tab to prevent comma from being removed.
                    // Tab is removed later.
                }
                if ( node.getDistanceToParent() >= 0.0 ) {
                    s += ( ":" + node.getDistanceToParent() );
                }
                A_stack.push( s );
                // Need to check whether parent is null in case the tree is only
                // one node:
                if ( !node.isRoot() ) {
                    node = node.getParent();
                }
                else {
                    node.setIndicator( 2 );
                }
            }
        } while ( !node.isRoot() || node.getIndicator() != 2 );

        //inverse:
        while ( !A_stack.empty() ) {
            B_stack.push( A_stack.pop() );
        }

        //make a nh string out of B_stack:
        nh_string = stackToString( B_stack );

        //In case tree is just one node -- return empty string:
        if ( nh_string.indexOf( "(" ) == -1 ) {
            return "";
        }

        return nh_string;
    }



    /**
  
    Converts this Tree to a New Hampshire X (String) representation.
    
    @return New Hampshire X (String) representation of this
    
    */
    public String toNewHampshireX() {
        if ( isEmpty() ) {
            return "";
        }

        int i = 0;
        String s = "", nh_string = "";
        Stack A_stack = new Stack();
        Stack B_stack = new Stack();
        Node node = root;

        setIndicatorsToZero();

        do {
            if ( !node.isExternal() ) {
                if ( node.getIndicator() == 0 ) {
                    node.setIndicator( 1 );
                    node = node.getChild1();
                }
                if ( node.getIndicator() == 1 ) {
                    node.setIndicator( 2 );
                    node = node.getChild2();
                }
                // Add internal node:
                if ( node.getIndicator() == 2 ) {

                    if ( !node.isPseudoNode()
                    || ( !allowMoreThanBinaryNodesInNHoutput()
                    && !node.getParent().isRoot() )
                    && ( !(node.getParent().getChild1().isPseudoNode()  &&
                    node.getParent().getChild2().isPseudoNode() ) ||
                    node.getParent().getChild1() == node ) ) {
                        i = 0;
                        while ( i < 2 * node.getSumExtNodes() - 2 ) {

                            B_stack.push( A_stack.pop() );
                            if (    !B_stack.peek().toString().equals( "(" )
                            &&  !B_stack.peek().toString().equals( ")" ) ) {
                                i++;
                            }
                        }
                        A_stack.push( "(" );
                        while ( !B_stack.empty() ) {
                            A_stack.push( B_stack.pop() );
                        }
                        A_stack.push( ")" );
                    }

                    // If there is info about an internal node, a "," needs to be added.
                    // In cases of ",)" the comma will be removed later.

                    s = "";
                    if ( !node.isPseudoNode() ) {

                        s = node.toNewHampshireX();
                        
                    }
                    else if ( node.isPseudoNode()
                    && !allowMoreThanBinaryNodesInNHoutput()
                    && !node.getParent().isRoot()
                    && ( !(node.getParent().getChild1().isPseudoNode()  &&
                    node.getParent().getChild2().isPseudoNode() ) ||
                    node.getParent().getChild1() == node ) ) {
                        s = ":0.0";
                    }
                    // If s contains something, add a comma:
                    if ( s.length() >= 1 ) {
                        s += ",";
                    }
                    A_stack.push( s );
                    node = node.getParent();
                }
            }
            // Add external node. Always preceeded by a "," which will need
            // to be removed in cases of "(,X"
            else {
                s = ",";
                s += node.toNewHampshireX();
               
                A_stack.push( s );

                // Need to check whether parent is null in case the tree is only
                // one node:
                if ( !node.isRoot() ) {
                    node = node.getParent();
                }
                else {
                    node.setIndicator( 2 );
                }
            }


        } while ( !node.isRoot() || node.getIndicator() != 2 );

        //inverse:
        while ( !A_stack.empty() ) {
            B_stack.push( A_stack.pop() );
        }
        //make a nh string out of B_stack:
        nh_string = stackToString( B_stack );

        // If information is present about the root, add it after the last ")":
        if ( getNumberOfExtNodes() >= 2
        && ( !root.getSeqName().equals( "" ) || root.getDistanceToParent() != 0.0
        || !root.getSpecies().equals( "" ) || !root.getECnumber().equals( "" )
        || root.isLnLonParentBranchAssigned() || root.isDuplicationOrSpecAssigned()
        || root.getBootstrap() != 0 ) ) {

            //remove ; at the end
            nh_string = nh_string.substring( 0, nh_string.length() - 1 );

            nh_string += node.toNewHampshireX();
            
            nh_string += ";";
        }

        return nh_string;

    }



    /**
  
    Private helper method used by the "toNewHampshire..." methods.

    */
    private String stackToString( Stack stack ) {
        int i;
        String s, string;
        StringBuffer sb;
        if ( getNumberOfExtNodes() >= 2 ) {
            string = "(";
        }
        else {
            // No need to add a "(" if tree has just one node
            string = "";
        }
        while ( !stack.empty() ) {
            s =  stack.pop().toString();
            //add "," in between ")(":
            if (  s.startsWith( "(" ) && string.endsWith( ")" )  ) {
                string += ( "," + s );
            }
            //add "," in between "'not(,'(":
            else if (  s.startsWith( "(" ) && !string.endsWith( "(" )
            && !string.endsWith( ","  )  ) {
                string += ( "," + s );
            }
            //get rid of the "," in "(,X":
            else if (  s.startsWith( "," ) && string.endsWith( "(" )  ) {
                string += s.substring( 1 );
            }
            //get rid of the "," in "X,)":
            else if (  s.startsWith( ")" ) && string.endsWith( "," )  ) {
                string = string.substring( 0, string.length() - 1 );
                string += s;
            }
            //get rid of one comma in ",,":
            else if (  s.startsWith( "," ) && string.endsWith( "," )  ) {
                string += s.substring( 1 );
            }
            else {
                string += s;
            }
        }

        // Remove "," at end of nh_string:
        if ( string.endsWith( "," )  ) {
            string = string.substring( 0, string.length() - 1 );
        }
        if ( getNumberOfExtNodes() >=2 ) {
            string += ");";
        }
        else {
            // No need to add a "(" if tree has just one node
            string += ";";
        }
        // Remove "," at begining of nh_string (can only happen
        // if tree contains just one node):
        if ( string.startsWith( "," )  ) {
            string = string.substring( 1 );
        }


        // Remove tabs which were introduced to protect commas originating from
        // unnamed ext nodes from being removed:
        sb = new StringBuffer( string );

        for ( i = 0; i <= sb.length() - 1; i++ ) {
            if ( sb.charAt( i ) == '\t' ) {
                sb = new StringBuffer( sb.toString().substring( 0, i ) +
                sb.toString().substring( i + 1 ) );
                i--;
            }
        }

        string = sb.toString();
        return string;
    }



    /**

    Returns a Vector containing the Node IDs of all Nodes which are between
    two external Nodes (node2 and node1).
    The order is so that element 0 is node2, and the last element is
    node1.
    Even if the Tree is rooted, this method does not return the root
    as part of the pathway (not ideal).
    The boolean ret_pseudo_nodes determines whether or not to return
    pseudo Nodes.
    (true: return pseudo nodes, false: only return real nodes.)
    
    */
    public Vector getPath( Node node1, Node node2, boolean ret_pseudo_nodes )
    throws Exception {

        if ( isEmpty() ) {
            return null;
        }

        if ( !node1.isExternal() || !node2.isExternal() ) {
            String message = "Tree: getPath( Node, Node ): ";
            message += "Nodes must be external.";
            throw new Exception( message );
        }

        Vector nodes = new Vector( 50, 50 );
        Tree tree = copyTree();

        if ( getNumberOfExtNodes() < 2 || node1 == node2 ) {
            nodes.addElement( new Integer( node1.getID() ) );
            return nodes;
        }

        try {
            tree.reRoot( node1.getID() );
            tree.adjustNodeCount();
            tree.recalculateAndReset();
        }
        catch ( Exception e ) {
            throw new Exception( "Tree: getPath( Node, Node ): " + e );
        }

        node1 = tree.getNode( node1.getID() );
        node2 = tree.getNode( node2.getID() );

        while ( !node2.isRoot() ) {
            if ( ret_pseudo_nodes || !getNode( node2.getID() ).isPseudoNode() ) {
                nodes.addElement( new Integer( node2.getID() ) );
            }
            node2 = node2.getParent();
        }

        nodes.addElement( new Integer( node1.getID() ) );

        nodes.trimToSize();

        return nodes;

    }



    /**

    Returns a Vector of references to all external siblings
    and nieces of a external Node n.
    The order is the same as in the Tree.

    @return Vector of references to Nodes

    */
    public Vector getSiblings( Node n ) throws Exception {

        if ( isEmpty() ) {
            return null;
        }

        if ( !n.isExternal() ) {
            String message = "Tree: getSiblings( Node ): ";
            message += "This method only works for external nodes.";
            throw new Exception( message );
        }

        if ( n.isRoot() ) {
            String message = "Tree: getSiblings( Node ): ";
            message += "Root has no sibling.";
            throw new Exception( message );
        }

        if ( n.isPseudoNode() ) {
            String message = "Tree: getSiblings( Node ): ";
            message += "Attempt to get sibling of a pseudonode.";
            throw new Exception( message );
        }


        Node node = n;
        Vector nodes;


        do {
            node = node.getParent();
        } while ( node.isPseudoNode() );

        nodes = node.getAllExternalChildren();
        // Since this methods guarantees that the order is not changed,
        // getAllExternalChildren() needs to return the Nodes in the same order
        // as in the Tree.

        // remove the original node n out of the vector:
        if ( !nodes.removeElement( n ) ) {
            String message = "Tree: getSiblings( Node ): ";
            message += "Unexpected failure.";
            throw new Exception( message );
        }

        nodes.trimToSize();

        return nodes;

    }



    /**

    Finds the last common ancestor Node of two Nodes specified by
    their sequence names seqname1 and seqname2. This method obviously
    can produce a meaningful result only, if this Tree is correctly
    rooted. Is inefficient and clumsy (O(n2)). Do not use.

    */
    public Node getLastCommonAncestor( String seqname1, String seqname2 )
    throws Exception {

        if ( isEmpty() ) {
            return null;
        }

        seqname1 = seqname1.trim();
        seqname2 = seqname2.trim();

        Vector nodes1 = getNodes( seqname1 ),
        nodes2 = getNodes( seqname2 );

        // Check whether names are unique and present:
        if ( nodes1 == null || nodes2 == null ) {
            String message = "Tree: getLastCommonAncestor( String, String ): ";
            message += "Node(s) which matching seq name(s) not found";
            throw new Exception( message );
        }
        if ( nodes1.size() != 1 || nodes2.size() != 1 ) {
            String message = "Tree: getLastCommonAncestor( String, String ): ";
            message += "Seq name(s) is (are) not unique.";
            throw new Exception( message );
        }


        int id1 = ( ( Node ) nodes1.firstElement() ).getID();
        int id2 = ( ( Node ) nodes2.firstElement() ).getID();

        return getLastCommonAncestor( id1, id2 );

    }



    /**

    Finds the last common ancestor Node of two Nodes specified by
    their IDs id1 and id2. This method obviously
    can produce a meaningful result only, if this Tree is correctly
    rooted. Is inefficient and clumsy (O(n2)). Do not use.

    */
    public Node getLastCommonAncestor( int id1, int id2 )
    throws Exception {

        if ( isEmpty() ) {
            return null;
        }
       
        Node node1 ,
        node2_,
        node2 = null;

        try {
            node1  = getNode( id1 );
            node2_ = getNode( id2 );
        }
        catch ( Exception e ) {
            String message = "Tree: getLastCommonAncestor( int, int ): ";
            message += "Node(s) not found.";
            throw new Exception( message );
        }


        // Find the "first common parent node":
        while ( !node1.isRoot() ) {

            node2 = node2_;

            while ( !node2.isRoot() ) {
                if ( node2 == node1 && !node2.isPseudoNode() ) {
                    return node2;
                }
                node2 = node2.getParent();
            }
            node1 = node1.getParent();

        }
        // In case the "first common parent node" is the root:
        if ( node2 == node1 && !node2.isPseudoNode() ) {
            return node2;
        }

        // If for some reason not found:
        throw new Exception ( "getLastCommonAncestor: Unexpected error." );
    }




    /**

    Adjusts the references (previous and next) in the
    external Nodes of this Tree.

    */
    private void adjustReferencesInExtNodes() {
        if ( isEmpty() ) {
            return;
        }

        // getIndicator() must be 0 for all nodes:
        getRoot().setIndicatorsToZero();

        Node node      = getRoot();
        Node prev_node = null;

        if ( getNumberOfExtNodes() <= 1 ) {
            return;
        }

        do {

            if ( node.getIndicator() == 0 && !node.isExternal() ) {
                node.setIndicator( 1 );
                node = node.getChild1();
            }

            if ( node.getIndicator() == 1 && !node.isExternal() ) {
                node.setIndicator( 2 );
                node = node.getChild2();
            }

            if ( node.isExternal() ) {

                node.setNextExtNode( null );
                node.setPrevExtNode( prev_node );
                if ( prev_node != null ) {
                    prev_node.setNextExtNode( node );
                }

                prev_node = node;
                node.setIndicator( 2 );

            }

            // Moving towards the root:
            if ( node.getIndicator() == 2 ) {
                node = node.getParent();
            }

        } while ( !node.isRoot() || node.getIndicator() != 2 );
    }



    /**
  
    Adjusts the "first" external Node.
   
    */
    private void adjustExtNode0() {
        Node node = getRoot();
        if ( node == null ) {
           return;
        }
        while ( !node.isExternal() ) {
            node = node.getChild1();
        }
        setExtNode0( node );
    }



    /**
  
    (Re)counts the number of children for each Node of this Tree.
    As an example, this method needs to be called after a Tree has
    been reRooted.
    (Last modifed 11/20/00)

    */
    public void adjustNodeCount() {
        if ( isEmpty() ) {
            return;
        }

        Node node1 = null,
             node2 = null;

        adjustNodeCountHelper( getRoot() );

        //Set all to the new, correct values:
        node2 = getExtNode0();
        O: do {
            node1 = node2;
            do {
                if ( node1.collapse() ) {
                    if ( node1.getIndicator() != -33 ) {
                        node1.setIndicator( -33 );  
                    }
                    else {
                        node2 = node2.getNextExtNode();
                        continue O;
                    }  
                }
                node1.setSumExtNodes( node1.getSumExtNodes() + 1 );
                node1 = node1.getParent();
            } while ( node1 != null );
            node2 = node2.getNextExtNode();
        } while( node2 != null );
    }



    private void adjustNodeCountHelper( Node n ) {
    
        if ( n == null ) {
            return;
        }
        else {
            n.setSumExtNodes( 0 );
            adjustNodeCountHelper( n.getChild1() );
            adjustNodeCountHelper( n.getChild2() );
        }
    }


    /**
  
    Reassigns the IDs of the Nodes of this Tree.
   
    */
    private void reassignIDs() throws Exception {
        if ( isEmpty() ) {
            return;
        }

        PreorderTreeIterator i = null;
        try {
            i = new PreorderTreeIterator( this );
        }
        catch ( Exception e ) {
            e.printStackTrace();
            System.err.println( "Tree: Unexpected failure." );
            System.exit( -1 );
        }
        while ( !i.isDone() ) {
            Node.setNodeCount( Node.getNodeCount() + 1 );
            i.currentNode().setID( Node.getNodeCount() );
            i.next();
        }
        idhash = null;
    }



    /**

    Used for building Trees.

    */
    private void addNode( String s ) throws Exception {
        if ( getExtNode0() == null ) {
            setExtNode0( new Node( s ) );
        }
        else {
            getExtNode0().addExtNode( s );
        }
        external_nodes++;
    }



    /**

    Used for building Trees.

    */
    void addNodeAndConnect( String s1, String s2 ) throws Exception {
        if ( getExtNode0() == null ) {
            throw new Exception( "addNodeAndConnect( String, String ): Error: Cannot add and connect one node to empty tree." );
        }
        getExtNode0().addExtNode( s1 );
        getExtNode0().connect( s2 );
        external_nodes++;
    }



    /**

    Used for building Trees.

    */
    private void connectInternal( String s ) throws Exception {
        if ( isEmpty() ) { return; }
        getExtNode0().connect( s );
    }


   
    /**

    Resets the ID numbers of the Nodes of this Tree in
    level order, starting with i (for the root).
    Warning. After this method has been called, node IDs
    are not longer unique.
    Implemented recursively.
    (Last modified: 11/20/00)
    
    @param i the starting value (int)

    */
    public void levelOrderReID( int i ) {
        idhash = null;
        levelOrderReIDHelper( getRoot(), i );
    }



    // Helper method for levelOrderReID.
    private void levelOrderReIDHelper( Node n, int label ) {
        n.setID( label );
        label++;
        if ( n.getChild1() != null ) {
            levelOrderReIDHelper( n.getChild1(), label );
        }
        if ( n.getChild2() != null ) {
            levelOrderReIDHelper( n.getChild2(), label );
        }
    }



    /**

    Resets the ID numbers of the Nodes of this Tree in
    preorder, starting with i. Implemented using a Stack.
    (Last modified: 11/20/00)    

    @param i the starting value (int)

    @return i plus the total number of Nodes of this Tree (int)
    
    */
    public int preorderReID( int i ) {

        if ( isEmpty() ) { 
            return i;
        }

        Stack stack = new Stack();
        Node n = getRoot();
        stack.push( n );
        while ( !stack.empty() ) {
            n = ( Node ) stack.pop();
            n.setID( i++ ); 
            if ( n.getChild2() != null ) {
                stack.push( n.getChild2() );
            }
            if ( n.getChild1() != null ) {
                stack.push( n.getChild1() );
            }
            
        }
        idhash = null;
        return i;
    }



    /**

    Hashes the ID number of each Node of this Tree to its corresonding
    Node, in order to make method getNode( id ) run in constant time.
    Important: The user is responsible for calling this method (again) after
    this Tree has been changed/created/renumbered. 

    */
    public void hashIDs() {
 
        idhash = null;

        if ( isEmpty() ) {
            return;
        }
             
        //idhash = new HashMap(); // Requires JDK 1.2 or greater.
        idhash = new Hashtable();  // For JDK 1.1.
        // Setting the initial capacity does not result in improved time efficiency.

        Stack stack = new Stack();
        Node n = getRoot();
        stack.push( n );
        while ( !stack.empty() ) {
            n = ( Node ) stack.pop(); 
            // ID is the key, reference to Node is the value.
            idhash.put( new Integer( n.getID() ), n );
            if ( n.getChild1() != null ) {
                stack.push( n.getChild1() );
            }
            if ( n.getChild2() != null ) {
                stack.push( n.getChild2() );
            }
        }
    }

} // End of class Tree.

